@@ -10,6 +10,7 @@
 #include <hwy/foreach_target.h>
 #include <hwy/highway.h>
 
+#include "lib/jxl/base/rect.h"
 #include "lib/jxl/convolve-inl.h"
 
 HWY_BEFORE_NAMESPACE();
@@ -232,16 +233,19 @@ class Separable5Strategy {
   }
 };
 
-void Separable5(const ImageF& in, const Rect& rect,
-                const WeightsSeparable5& weights, ThreadPool* pool,
-                ImageF* out) {
+Status Separable5(const ImageF& in, const Rect& rect,
+                  const WeightsSeparable5& weights, ThreadPool* pool,
+                  ImageF* out) {
   using Conv = ConvolveT<Separable5Strategy>;
   if (rect.xsize() >= Conv::MinWidth()) {
+    JXL_ENSURE(SameSize(rect, *out));
+    JXL_ENSURE(rect.xsize() >= Conv::MinWidth());
+
     Conv::Run(in, rect, weights, pool, out);
-    return;
+    return true;
   }
 
-  SlowSeparable5(in, rect, weights, pool, out, Rect(*out));
+  return SlowSeparable5(in, rect, weights, pool, out, Rect(*out));
 }
 
 // NOLINTNEXTLINE(google-readability-namespace-comments)
@@ -253,10 +257,10 @@ HWY_AFTER_NAMESPACE();
 namespace jxl {
 
 HWY_EXPORT(Separable5);
-void Separable5(const ImageF& in, const Rect& rect,
-                const WeightsSeparable5& weights, ThreadPool* pool,
-                ImageF* out) {
-  HWY_DYNAMIC_DISPATCH(Separable5)(in, rect, weights, pool, out);
+Status Separable5(const ImageF& in, const Rect& rect,
+                  const WeightsSeparable5& weights, ThreadPool* pool,
+                  ImageF* out) {
+  return HWY_DYNAMIC_DISPATCH(Separable5)(in, rect, weights, pool, out);
 }
 
 }  // namespace jxl
